/*
 * bitmap.c                (C) 2004-2009 Dmitry Yakovlev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
                            
#include <stdio.h>
#include <string.h>
#include "qnx4fs_fs.h"
                                   
static inline int test_bit(int nr, const volatile unsigned char *addr) {
    return ((1 << (nr & 7)) & (addr[nr >> 3])) != 0;
}

static inline int test_and_set_bit(int nr, volatile unsigned char *addr) {
	if(test_bit(nr, addr)) return 1;
	addr[nr >> 3] |= 1 << (nr & 7);
	return 0;
}

static inline int check_bits (u8 *ptr, int from, int len) {
	for (; len; len--)
		if(!test_bit(from++, (unsigned char*)ptr)) 
			return chkfs_EBIT_NSET;
	return 0;
}

static inline int check_low_bits (u8 *ptr, int len) {
	if((*ptr & (~(0xff << len))) != ((~(0xff << len)) & 0xff)) return chkfs_EBIT_NSET;
	return 0;
}

static inline int test_and_set_bits (u8 *ptr, int from, int len) {
	int result = 0;
	for (; len; len--)
		if(test_and_set_bit(from++, (unsigned char*)ptr)) 
			result |= chkfs_EBIT_SET;
	return result;
}

static inline int test_and_set_low_bits (u8 *ptr, int len) {
	int result = ((*ptr & (~(0xff << len))) != 0) ? chkfs_EBIT_SET : 0;
	*ptr |= ~(0xff << len);
	return result;
}

static inline int check_high_bits (u8 *ptr, int len) {
	if((*ptr & ~(0xff >> len)) != (~(0xff >> len) & 0xff)) return chkfs_EBIT_NSET;
	return 0;
}

static inline int test_and_set_high_bits (u8 *ptr, int len) {
	int result = (((*ptr & (~(0xff >> len))) & 0xff) != 0) ? chkfs_EBIT_SET : 0;
	*ptr |= ~(0xff >> len);
	return result;
}

static inline int check_bytes (u8 *bitmap, int from, int to) {
	for(bitmap += from, to -= from; to > 0; to--)
		if(*bitmap++ != 0xff) return chkfs_EBIT_NSET;
	return 0;
}

static inline int test_and_set_bytes (u8 *bitmap, int from, int to) {
	int result = 0;
	for(bitmap += from, to -= from; to > 0; to--) {
		if(*bitmap != 0) result |= chkfs_EBIT_SET;
		*bitmap++ = 0xff;
	}
	return result;
}

static inline struct buffer_head *
bitmap_sb_bread (struct super_block *sb, unsigned long block) {
	struct buffer_head *bh;
	bh = sb_bread(sb, block + QNX4FS_BITMAP_BLOCK - 1);
	if(!bh) fprintf(stderr," Unable to read bitmap block %lu\n", block);
	return	bh;
}

static inline int
check_sb_blocks (struct super_block *sb, unsigned long from, unsigned long to) {
	struct buffer_head *bp;
	unsigned char *bpt;
	int i;

	for (; from < to; from++) {
		bp = bitmap_sb_bread(sb, from);
		if (bp == NULL)
			continue;
		bpt = (unsigned char*)(bp->b_data);
		for(i = 0; i < QNX4FS_BLOCK_SIZE; i++){
			if(*bpt++ != 0xFF){
				brelse(bp);
				return chkfs_EBIT_NSET;
			}
		}
		brelse(bp);
	}
	return 0;
}

int qnx4fs_bcheck (struct super_block *sbp, unsigned long from, unsigned long len)
{
	struct buffer_head *bp;
	unsigned long to;
	unsigned int bfrom, bto;
	unsigned long block_from, block_to;
	u8 *bitmap;
	int res = 0;

	if (len == 0) return 0;
	to = --from + len;

	if (to > QNX4FS_SB(sbp)->s_blocks){
		fprintf(stderr," Block range error: from %lu, len %lu, to %lu\n", from + 1, len, to);
		return -1;
	}

	block_from = from >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	block_to   = to >> (QNX4FS_BLOCK_SIZE_BITS + 3);
	from      &= (QNX4FS_BLOCK_SIZE * 8 - 1);
	to        &= (QNX4FS_BLOCK_SIZE * 8 - 1);
	bfrom      = from >> 3;
	bto        = to >> 3;

	bp = bitmap_sb_bread(sbp, block_to);
	if (bp == NULL)
		return -1;
	bitmap = (u8 *)bp->b_data;
	switch (block_to - block_from) {
	case 0:
		switch (bto - bfrom) {
		case 0:
			res |= check_bits(&bitmap[bfrom], from & 7, len);
			break;
		default:
			res |= check_low_bits(&bitmap[bto], to & 7);
			res |= check_bytes(bitmap, (from + 7) >> 3, bto);
			res |= check_high_bits(&bitmap[bfrom], 8 - (from & 7));
			break;
		}
		break;
	default:
		res |= check_low_bits(&bitmap[bto], to & 7);
		res |= check_bytes(bitmap, 0, bto);
		brelse(bp);
		res |= check_sb_blocks(sbp, block_from + 1, block_to);
		bp = bitmap_sb_bread(sbp, block_from);
		if (bp == NULL)
			return -1;
		bitmap = (u8 *)bp->b_data;
		res |= check_bytes(bitmap, (from + 7) >> 3, QNX4FS_BLOCK_SIZE);
		res |= check_high_bits(&bitmap[bfrom], 8 - (from & 7));
		break;
	}
	brelse(bp);
	return res;
}

int qnx4fs_btest_and_set (struct super_block *sbp, unsigned long from, unsigned long len) {
	unsigned long to;
	unsigned int bfrom, bto;
	u8 *bitmap;
	int result = 0;

	if (len == 0) return 0;
	to = --from + len;

	if (to > (QNX4FS_SB(sbp)->s_bitmap_blocks << 12)) {
		fprintf(stderr," block range error: from %lu, len %lu, to %lu\n", from + 1, len, to);
		return -1;
	}

	bfrom      = from >> 3;
	bto        = to >> 3;

	bitmap = QNX4FS_SB(sbp)->s_bitmap_new;
	if((bto - bfrom) == 0) 
		result |= test_and_set_bits(&bitmap[bfrom], from & 7, len);
	else {
		result |= test_and_set_low_bits(&bitmap[bto], to & 7);
		result |= test_and_set_bytes(bitmap, (from + 7) >> 3, bto);
		if(from & 7)
		result |= test_and_set_high_bits(&bitmap[bfrom], 8 - (from & 7));
	}
	return result;
}

int qnx4fs_check_bitmap (struct super_block *sbp, int verbose)
{
	struct buffer_head *bp;
	unsigned long nblocks, block;
	unsigned char *nbm;
	int i;
	int result = 0;

	nblocks = QNX4FS_SB(sbp)->s_bitmap_blocks;
	nbm     = QNX4FS_SB(sbp)->s_bitmap_new;
	for (block = 0; block < nblocks; block++, nbm += QNX4FS_BLOCK_SIZE) {
		bp = bitmap_sb_bread (sbp, block);
		if (bp == NULL)
			continue;
		for(i = 0; i < QNX4FS_BLOCK_SIZE; i++){
			if(bp->b_data[i] != nbm[i]){
				if(verbose)
				fprintf(stderr, "Fs block %7lu bitmap block %5lu, byte %3d: %2x, should be %2x\n",
				(block * 512 + i) * 8 , block, i, bp->b_data[i], nbm[i]);
				result += generic_hweight8((unsigned) (bp->b_data[i] ^ nbm[i]));
			}
		}
		brelse(bp);
	}
	if(result) {
		fprintf(stderr, " Bitmap check detected %d different bits.\n", result);
		return result;
	}
	else {
		printf(" New bitmap is equal to old bitmap.\n");
	}
	return 0;
}

int qnx4fs_write_new_bitmap (struct super_block *sbp) {
	struct buffer_head *bp;
	unsigned long nblocks, block;
	unsigned char *nbm;

	nblocks = QNX4FS_SB(sbp)->s_bitmap_blocks;
	nbm     = QNX4FS_SB(sbp)->s_bitmap_new;
	for (block = 0; block < nblocks; block++, nbm += QNX4FS_BLOCK_SIZE) {
		bp = bitmap_sb_bread (sbp, block);
		if (bp == NULL) continue;
		if(memcmp(bp->b_data, nbm, QNX4FS_BLOCK_SIZE)){
			memcpy(bp->b_data, nbm, QNX4FS_BLOCK_SIZE);
			chkfs_bwrite(sbp, bp);
		}
		brelse(bp);
	}
	return 0;
}


