/*
 *  check.c     Copyright (C) 2004-2009  Dmitry Yakovlev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "qnx4fs_fs.h"

#define MAX_LEVEL 100
static int level;

#define HASH_SIZE 2048

#define hashfn(x) \
    ((((x) >> 12) ^ (x)) & (HASH_SIZE - 1))

struct chkfs_LINODE {
	struct chkfs_LINODE *next;
	ino_t  ino;
	u16    nlink, links, mode, spare;
};

static struct chkfs_LINODE *hashtable[HASH_SIZE];

#define PACKSIZE 1000

struct chkfs_LINODE *freelist = NULL;

static inline struct chkfs_LINODE* chkfs_get_new() {
	struct chkfs_LINODE *next;
	int i;
	
	if(!freelist){
		next = malloc(sizeof(*next) * PACKSIZE);
		if(!next) {
			perror(" inodes hash buffer malloc");
			exit(EXIT_FAILURE);
		}
		for(i = 0; i < PACKSIZE; i++, next++){
			next->next = freelist;
			freelist = next;
		}
	}
	next = freelist;
	freelist = next->next;
	return next;
}

static int chkfs_inode_checked(struct inode *ip){
	struct chkfs_LINODE *next, **list;

	list = &hashtable[hashfn(ip->i_ino)];
	next = *list;
	while(next){
		if(next->ino == ip->i_ino){
			next->links++;
			return 1;
		}
		next = next->next;
	}
	next = chkfs_get_new();
	next->ino   = ip->i_ino;
	next->nlink = ip->i_nlink;
	next->mode  = ip->i_mode;
	next->links = 1;
	next->next = *list;
	*list = next;
	return 0;
}

void chkfs_check_links (struct super_block* sb) {
	struct chkfs_LINODE *next;
	struct inode dinode;
	int i;
	
	for(i = 0; i < HASH_SIZE; i++) {
		next = hashtable[i];
		while(next) {
			if(!S_ISDIR(next->mode) && (next->nlink != next->links)){
				fprintf(stderr,"\n wrong inode %u nlink field"
				": %u, should be %u\n",	(unsigned)next->ino, 
				next->nlink, next->links);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT){
					dinode.i_ino = next->ino;
					dinode.i_sb  = sb;
					if(qnx4fs_read_inode(&dinode) < 0);
					else {
						chkfs_edit_inode(&dinode);
					}
				}
			}
			next = next->next;
		}
	}
}

int chkfs_check_dots_entries(struct inode *ip, struct inode *dip){
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	ino_t ino;
	extern char verbose;

	bp = sb_bread(ip->i_sb, QNX4FS_I(ip)->i_xt0.start - 1);

	if (bp == NULL){
		/* we checked extent range in qnx4fs_check_file() */
		fprintf(stderr," This should not happen!\n"
		" Can't read directory first extent\n");
		return -1;
	}

	qdp = (qnx4fs_dirent_t *)bp->b_data;

	if (qdp->i.status != QNX4FS_FILE_LINK) {
		fprintf(stderr,"\n <%s>\n directory dot entry wrong status"
		" %02x\n", chkfs_get_name(), qdp->i.status);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
		goto quitbp;
	}

	if(strncmp((char*)bp->b_data, ".", 2) != 0){
		fprintf(stderr, "\n <%s>\n warning: wrong directory dot entry name"
		" (ino %u).\n", chkfs_get_name(), (unsigned)ip->i_ino);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
	}

	if(memcmp((char*)bp->b_data + 2, "I\03QNX\0", 6)){
		fprintf(stderr,"\n <%s>\n wrong directory dot entry signature\n",
		chkfs_get_name());
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
	}

	if(((struct qnx4fs_dot_entry*)bp->b_data)->block != QNX4FS_I(ip)->i_xt0.size){
		fprintf(stderr,"\n <%s>\n wrong directory dot entry extent length number"
		": %u, should be %u\n",	chkfs_get_name(), 
		le32_to_cpu(((struct qnx4fs_dot_entry*)bp->b_data)->block),
		QNX4FS_I(ip)->i_xt0.size);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
	}

	ino = link2ino(qdp);
	if(ino != ip->i_ino){
		fprintf(stderr, "\n <%s>\n warning: wrong directory dot entry link"
		" ino %u, should be %u(%lu,%u).\n",
		chkfs_get_name(), (unsigned)ino, (unsigned)ip->i_ino,
		ino2phys(ip->i_ino) + 1, ino2slot(ip->i_ino));
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
	}

	qdp++;

	if (qdp->i.status != QNX4FS_FILE_LINK) {
		fprintf(stderr,"\n <%s>\n directory dot dot entry wrong status %02x\n", 
		chkfs_get_name(), qdp->i.status);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
		goto quitbp;
	}

	if(strcmp((char*)qdp, "..") != 0){
		fprintf(stderr, "\n <%s>\n wrong directory dot dot entry name, ino %u\n",
		chkfs_get_name(), (unsigned)ip->i_ino);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_dir(ip);
	}

	ino = link2ino(qdp);

	if(verbose)
	if((ino != dip->i_ino) && (ino != direct2ino(QNX4FS_I(dip)->i_xt0.start, 0))){
		fprintf(stderr, "\n <%s>\n warning: wrong directory dot dot entry "
		"link ino %u, should be %u or %u\n", 
		chkfs_get_name(), (unsigned)ino, (unsigned)dip->i_ino,
		(unsigned) direct2ino(QNX4FS_I(dip)->i_xt0.start, 0));
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
		chkfs_edit_dir(ip);
	}
	brelse (bp);
	return 0;

quitbp:
	brelse (bp);
	return -1;
}

void chkfs_check_dir(struct inode *ip, struct inode *dip){
	struct qnx4fs_dir dir;
	struct inode dinode;
	char *str;

	printf("%s\n",  chkfs_get_name());
	memset(&dir, 0, sizeof(dir));
	chkfs_check_dots_entries(ip, dip);
	/* skip dots entries, which were checked above */
	dir.d_count = 2;

	while(qnx4fs_readdir(ip, &dir)){
		int res;

		dir.d_count++;
		memset(&dinode, 0, sizeof(dinode));
		
		dinode.i_ino = dir.d_ino;
		dinode.i_sb  = ip->i_sb;

		str = chkfs_add_name('/', dir.d_name, sizeof(dir.d_name));

		if(qnx4fs_read_inode(&dinode) < 0){

			fprintf(stderr,"\n <%s>\n failed to read ino %u, block %u, slot %u\n", 
			chkfs_get_name(), (unsigned)dinode.i_ino, (unsigned)ino2phys(dinode.i_ino),
			(unsigned)ino2slot(dinode.i_ino));
			
			if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
				chkfs_edit_dir (ip);

			goto next_entry;
		}
		
		if(chkfs_inode_checked(&dinode)) {
			if(S_ISDIR(dinode.i_mode)){
				fprintf(stderr, "\n <%s>\n directory inode %u already checked, remove the entry\n",
				chkfs_get_name(), (unsigned)dinode.i_ino);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
					chkfs_edit_dir (ip);
			}
			goto next_entry;
		}

		if((res = qnx4fs_check_file(&dinode)) != 0){
			if(res > 0){
				if(res & chkfs_EBIT_NSET){
					fprintf(stderr, "\n <%s>\n some blocks marked free\n", 
					chkfs_get_name());
					if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
						chkfs_edit_inode(ip);
				}
				if(res & chkfs_EBIT_SET){
					fprintf(stderr, "\n <%s>\n some blocks marked used by another file\n",
					chkfs_get_name());
					if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
						chkfs_edit_inode(ip);
				}
			}
			goto next_entry;
		}
		
		if(S_ISDIR(dinode.i_mode)){
			if(++level > MAX_LEVEL) {
				fprintf(stderr, " too deep\n");
				chkfs_press_keym(0);
				break;
			}
			chkfs_check_dir(&dinode, ip);
			level--;
		}
next_entry:
		chkfs_restore_name(str);
	}
}

void chkfs_copy_dir(struct inode *ip){
	struct qnx4fs_dir dir;
	struct inode dinode;
	char *str;

	printf("%s\n",  chkfs_get_name());
	memset(&dir, 0, sizeof(dir));
	dir.d_count = 2;

	while(qnx4fs_readdir(ip, &dir)){
		int res;

		dir.d_count++;
		memset(&dinode, 0, sizeof(dinode));
		
		dinode.i_ino = dir.d_ino;
		dinode.i_sb  = ip->i_sb;
		str = chkfs_add_name('/', dir.d_name, sizeof(dir.d_name));

		if(qnx4fs_read_inode(&dinode) < 0) {
			fprintf(stderr,"\n <%s>\n failed to read inode %u (block %u, slot %u)\n", 
			chkfs_get_name(), (unsigned)dinode.i_ino, (unsigned)ino2phys(dinode.i_ino),
			(unsigned)ino2slot(dinode.i_ino));
			
			if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			    chkfs_edit_dir (ip);
			goto next_entry;
		}

		if(chkfs_inode_checked(&dinode)) {
			if(S_ISDIR(dinode.i_mode)){
				fprintf(stderr, "\n <%s>\n directory inode %u already copied, remove the entry\n",
				chkfs_get_name(), (unsigned)dinode.i_ino);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
					chkfs_edit_dir (ip);
			}
			goto next_entry;
		}

		if((res = qnx4fs_check_file(&dinode)) != 0){
			if(res > 0){
				if(res & chkfs_EBIT_NSET){
					fprintf(stderr, "\n <%s>\n Some blocks marked free\n", 
					chkfs_get_name());
					if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
						chkfs_edit_inode(ip);
				}
				if(res & chkfs_EBIT_SET){
					fprintf(stderr, "\n <%s>\n some blocks marked used by another file\n",
					chkfs_get_name());
					if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
						chkfs_edit_inode(ip);
				}
			}
			goto next_entry;
		}

		if(S_ISREG(dinode.i_mode)){
			int fd;
#ifdef WIN32
			fd = open(chkfs_get_name(), O_CREAT | O_WRONLY | O_BINARY);
#else
			fd = creat(chkfs_get_name(), 0777 & dinode.i_mode);
#endif
			if(fd < 0) {
				fprintf(stderr, " can't create file <%s>", chkfs_get_name());
				perror("");
				goto next_entry;
			}

			qnx4fs_read_file(&dinode, fd);
			close(fd);
		} else

		if(S_ISDIR(dinode.i_mode)){
			if(++level > MAX_LEVEL) {
				fprintf(stderr, " too deep\n");
				chkfs_press_keym(0);
				break;
			}
			if(!mkdir(chkfs_get_name(), 0777)) {
				chkfs_copy_dir(&dinode);
			}
			else {
				fprintf(stderr, " can't make dir <%s>", chkfs_get_name());
				perror(" ");
			}
			level--;
		}
next_entry:
		chkfs_restore_name(str);
	}
}

struct inode *chkfs_namei(struct super_block* sb, const char *name){
	struct qnx4fs_dir dir;
	struct inode *ip;
	const char *str;
	static struct inode dinode;
	int res;

	str = name;
	dinode.i_ino = QNX4FS_ROOT_INO;
	dinode.i_sb = sb;
	for(ip = &dinode, res = 1;res;){

		if(qnx4fs_read_inode(ip) < 0) return NULL;
		if(*str == '/') str++;
		if(!*str) return ip;
		memset(&dir, 0, sizeof(dir));
		if(!S_ISDIR(ip->i_mode)) return NULL;

		for(;(res = qnx4fs_readdir(ip, &dir)) != 0; dir.d_count++){
			if(strncmp(dir.d_name, str, dir.d_namelen) == 0){
				if(str[dir.d_namelen] != '/' 
				&& str[dir.d_namelen] != 0)
					continue;
				str += dir.d_namelen;
				ip->i_ino = dir.d_ino;
				break;
			}
		}
	}
	return NULL;
}

void chkfs_dump_area(const void *buf, int sze){
	const unsigned char* bf = buf;
	int i, j;

	sze &= 0xFFF0;
	for(i = 0; i < sze; i += 16){

		printf(" %03X:", i);
		for(j = 0; j < 16; j++){
			if(j == 8) 	printf("  %02x", bf[i + j]);
			else if((j & 3) == 3)
					printf(" %02x ", bf[i + j]);
			else   		printf(" %02x",  bf[i + j]);
		}
		printf("  ");

		for(j = 0; j < 16; j++){
			if((bf[i + j] <= 'z') && (bf[i + j] >= 0x20))
				printf("%c", bf[i + j]);
			else
				printf("_");
		}
		printf("\n");
	}
}

void chkfs_dump(struct super_block* sb, unsigned long block){
	struct buffer_head *bh = sb_bread(sb, block - 1);
	if(bh == NULL) return;

	chkfs_dump_area(bh->b_data, QNX4FS_BLOCK_SIZE);
	brelse(bh);
	printf("\n");
}

int chkfs_check_finodes (struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long nblocks, block;
	int slot;
	ino_t ino;
	struct inode inode;
	int res = 0;

	nblocks = qnx4fs_nblocks(ip);

	for (block = 0; block < nblocks; block++) {

		phys = qnx4fs_map_block(ip, block);
		if (!phys) {
			fprintf(stderr, "\n can't map block %ld of /.inodes file\n", block);
			res |= -1;
			continue;
		}
		bp = sb_bread (ip->i_sb, phys - 1);
		if (bp == NULL) {
			fprintf(stderr, "\n can't read block %ld of /.inodes file\n", block);
			res |= -1;
			continue;
		}
		slot = 0;
		qdp = (qnx4fs_dirent_t *)bp->b_data; //  + slot;

		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {

			if (!qdp->i.status)
				continue;

			ino = direct2ino(phys, slot);
			inode.i_ino = ino;
			inode.i_sb  = ip->i_sb;

			if ((qdp->i.status & (QNX4FS_FILE_INODE | QNX4FS_FILE_USED))
			!= (QNX4FS_FILE_INODE | QNX4FS_FILE_USED)) {
				res |= -1;

				fprintf(stderr, "\n Wrong status field %02x of inode %u in /.inodes\n", 
					qdp->i.status, (unsigned)ino);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
					chkfs_edit_inode(&inode);
				continue;
			}			

			if(!chkfs_inode_checked(&inode)) {
				res |= -1;

				fprintf(stderr, "\n Zombie inode %u in /.inodes\n", 
					(unsigned)ino);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
					chkfs_edit_inode(&inode);
			}			
		}
		brelse(bp);
	}

	return res;
}

