/*
 *  chkfs.c     Copyright (C) 2004-2009  Dmitry Yakovlev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

static char sms[] = "chkfs V0.3b 21 apr 2009 Y.D. (http://qnxfs.narod.ru)";

static char usage[] =
"Usage:\n\n"
" chkfs  [-xlqrve] [-n file] [-i ino] [-d blk] [-f name] [-o file] [-O offset] drive\n"
"Options:\n"
" -d blk  Dump block blk.\n"
" -i ino  Display internal file information.\n"
" -n file Display internal file information.\n"
" -l      When -i, -d or -n keys set - interpret file (block)\n"
"         as a directory and list it contents.\n"
" -e      edit directory entries, inode,  extent-blocks, blocks\n"
" -r      When -i or -n keys set - read file to local dir,\n"
"         interpret any file (directory, soft link) as regular file.\n"
" -O ofs  Offset in blocks from disk start.\n"
" -S sze  Force device size (blocks).\n"
" -R      When -i or -n keys set - recursive files reading.\n"
" -o file When -r or -R keys set - output file name.\n"
" -f name Find directory entry \"name\" on disk.\n"
" -q      Quiet display.\n"
" -v      Verbose output.\n"
"Where:\n"
" drive is the disk, as:  /dev/fd0, /dev/hda2, /dev/hd0.0t77 etc.\n\n"
"For example:\n"
"\n"
"Check file system on /dev/hda2:\n"
"#chkfs  /dev/hda2\n"
"\n"
"Display root block content:\n"
"#chkfs -d 2 -l /dev/hda2\n"
"\n"
"List root directory:\n"
"#chkfs -n / -l /dev/hda2 | less\n"
"\n"
"Edit dir entries of root directory:\n"
"#chkfs -eln / /dev/hda2\n"
"\n"
"Edit dir entries of root block:\n"
"#chkfs -eld 2 /dev/hda2\n"
"\n"
"Edit inode of /.inodes file:\n"
"#chkfs -ei 9 /dev/hda2\n"
"\n"
"Display internal structure of /.inodes file:\n"
"#chkfs -i 9 /dev/hda2\n"
"\n"
"Find \"home\" dir entry:\n"
"#chkfs -f home /dev/hda2\n"
"\n"
"Read file /usr/bin/gzip to ./gzip:\n"
"#chkfs -n /usr/bin/gzip -r /dev/hda2\n"
#if defined(WIN32)
"\n"
"For Win32(NT/2000/XP) port use \\\\?\\a: as a device name of floppy a:.\n"
"C:>chkfs -v \\\\?\\a:\n"
#endif
"\n";

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include "qnx4fs_fs.h"

static struct super_block sb;
static char fname [1000];
static char ofile [1000];
static char dev   [1000];
static unsigned long dump;
static char  readf, rreadf, quiet, find, edit, list, exitx;
char verbose;
static ino_t ino = 0;

int chkfs_press_keym (int mode){
	char str[7] = {0};
	if(exitx) exit(EXIT_SUCCESS);
	if(quiet == 0) {
		int i;
		if(mode & chkfs_EDIT)
			fprintf(stderr, " [e - edit] ");
		if(mode & chkfs_YES)
			fprintf(stderr, " [y/n] ");
		fprintf(stderr, "press enter... ");
		fflush(stdout);
		fgets(str, sizeof(str), stdin);
		for(i = 0; i < sizeof(str); i++)
			if(str[i] != ' ') break;
		if(str[i] == 'e') return chkfs_EDIT;
		if(str[i] == 'y') return chkfs_YES;
	}
	return 0;
}

int chkfs_find_dentry(struct super_block *sb, char* fname){
	unsigned long i, res = 0;

	if(qnx4fs_get_devsize (sb)){
		fprintf(stderr, " Can't get device size\n");
		return -1;
	}

	printf(" Lookup <%s> entry\n", fname);
	for(i = 1; i <= QNX4FS_SB(sb)->s_blocks; i++)
		res += qnx4fs_find_block (sb, i, fname);
	if(!res)
		printf(" Entry <%s> not found\n", fname);
	return 0;
}

int chkfs_check_filesystem(){
	struct inode *ip, dinode;
	int res;

	if(qnx4fs_check_super(&sb)) return -1;

	dinode.i_ino = QNX4FS_LOST_AND_FOUND_INO;
	dinode.i_sb = &sb;
	ip = &dinode;


	// ???
	chkfs_set_name("/lost+found", 100);
	if(chkfs_namei(&sb, "/lost+found") == NULL)
	if(qnx4fs_read_inode(ip) == 0)
	if((res = qnx4fs_check_file(ip)) != 0)
		return -1;

	chkfs_set_name("/", 100);
	if((res = qnx4fs_check_file(QNX4FS_SB(&sb)->s_root_dir)) != 0)
		return -1;
	chkfs_check_dir(QNX4FS_SB(&sb)->s_root_dir, QNX4FS_SB(&sb)->s_root_dir);
	chkfs_check_links (&sb);

	ip->i_ino = QNX4FS_INODES_INO;
	qnx4fs_read_inode(ip);
	chkfs_check_finodes(ip);
		// printf(" File ./inodes OK\n");

	if(qnx4fs_check_bitmap (&sb, verbose) && !quiet){
		fprintf(stderr, " Write new bitmap? ");
		if(chkfs_press_keym(chkfs_YES) == chkfs_YES)
			qnx4fs_write_new_bitmap (&sb);
	}
	
	return 0;
}

char* chkfs_parse_name(){
	char *str = NULL;

	if(*fname) {
		str = fname;
		while(strchr(str, '/'))
			str = strchr(str, '/') + 1;
	}

	if(*ofile) 
		str = ofile;
	
	if(str == NULL || *str == 0){
		sprintf(fname, "ino_%u", (unsigned)ino);
		str = fname;
	}
	return str;
}

int chkfs_rread_file(struct inode *ip){
	chkfs_set_name(chkfs_parse_name(), 100);

	if(qnx4fs_get_devsize (&sb)){
		fprintf(stderr, " Can't get device size\n");
		return -1;
	}

	if(S_ISDIR(ip->i_mode)){
		if(mkdir(chkfs_get_name(), 0777)){
			fprintf(stderr, " rread: can't make dir <%s>", chkfs_get_name());
			perror(" ");
			return -1;
		}
		chkfs_copy_dir(ip);
		return 0;
	}
	return -1;
}

int chkfs_read_file(struct inode *ip){
	int fd;
	char *str = chkfs_parse_name();

	printf(" Writing to ./%s ...", str);
	fflush(stdout);
#ifdef WIN32
	fd = open(str, O_CREAT | O_WRONLY | O_BINARY);
#else
	fd = creat(str, 0777 & ip->i_mode);
#endif
	if(fd < 0) {
		perror("unable to create output file");
		return -1;
	}
	if(qnx4fs_read_file(ip, fd) == 0)
		printf(" OK\n");
	close(sb.s_fd);
	close(fd);
	return 0;
}

static int parse_options(int argc, char** argv){
	int i;

	if(argc <= 1) return -1;

	for(;;){

		i = getopt (argc, argv, "lxerRvqi:d:n:f:o:O:S:");

		if (i == -1) break;

		switch (i){
		case 'l': list    = 1; break;
		case 'q': quiet   = 1; break;
		case 'x': exitx   = 1; break;
		case 'v': verbose = 1; break;
		case 'e': edit    = 1; break;
		case 'r': readf   = 1; break;
		case 'R': rreadf  = 1; break;
		case 'o':
			strncpy(ofile, optarg, sizeof(ofile) - 1);
			break;
		case 'n':
			strncpy(fname, optarg, sizeof(fname) - 1);
			break;
		case 'f':
			find = 1;
			strncpy(fname, optarg, sizeof(fname) - 1);
			break;
		case 'i':
			ino = atoi(optarg);
			break;

		case 'd':
			dump = atoi(optarg);
			break;
		case 'O':
			sb.s_blocks_offset = atoi(optarg);
			break;
		case 'S':
			QNX4FS_SB(&sb)->s_blocks = atoi(optarg);
			break;
		case '?':
		default:
			return -1;
		}
	}

	if (optind < argc)
		strncpy(dev, argv[optind], sizeof(dev) - 1);

	if(dev[0] == 0) return -1;

	return 0;
}

int main(int argc, char** argv){
	struct inode *ip, dinode;

	if(parse_options(argc, argv) < 0) goto wrongkey;

	ip = &dinode;

#ifdef WIN32
	sb.s_fd = open(dev, O_RDONLY | O_BINARY);
#else
	sb.s_fd = open(dev, O_RDONLY);
#endif

	if(sb.s_fd < 0) {
		fprintf(stderr, " unable to open device <%s>: ", dev);
		perror("");
		return -1;
	}
	sb.s_devname = dev;

	if(dump) {
		if(list) {
			if(edit) 
				chkfs_edit_dir_blk(&sb, dump);
			else
				qnx4fs_lsblock (&sb, dump, 1);
		}
		else {

			if(edit) 
				chkfs_edit_block(&sb, dump);
			else
				chkfs_dump(&sb, dump);
		}
		return 0;
	}

	if(*fname) {

		if(find)
			return chkfs_find_dentry(&sb, fname);

		ip = chkfs_namei(&sb, fname);
		if(!ip) goto notfound;

		ino = ip->i_ino;

		if(memcmp(fname, "/", 2))
			chkfs_set_name(fname, 100);
	}

	if(ino){
		dinode.i_ino = ino;
		dinode.i_sb = &sb;
		ip = &dinode;

		if(edit && !list) {
			chkfs_edit_inode(ip);
			return 0;
		}

		if(qnx4fs_read_inode(ip) < 0)
			return 0;

		if(list) {
			if(edit) {
				chkfs_edit_dir(ip);
				return 0;
			}
			qnx4fs_print_inode(ip);
			printf(" list:\n");
			qnx4fs_lsdir (ip);
			return 0;
		}

		if(rreadf) {
			return chkfs_rread_file(ip);
		}

		if(readf) 
			return chkfs_read_file(ip);

		qnx4fs_print_inode(ip);
		qnx4fs_print_xblocks(ip);
	}
	else {
		chkfs_check_filesystem();
	}
	return 0;

notfound:
	fprintf(stderr, "file <%s> not found\n", fname);
	return 0;

wrongkey:
	printf("\n      %s\n\n", sms);
	printf(usage);
	return 0;
}
