/*
 * dir.c                   Copyright (C) 2004-2009 Dmitry Yakovlev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */
                                      
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include "qnx4fs_fs.h"

static char* dformat = 
"    inode stat    size  nbxts  xtnt0  xtnt0sze \n";

int qnx4fs_lsblock (struct super_block* sb, unsigned long block, int pheader)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	int slot, namelen;
	ino_t ino;
	struct qnx4fs_dir dir, *dirnt;

	dirnt = &dir;
	memset(&dir, 0, sizeof(dir));

	bp = sb_bread (sb, block - 1);
	if (bp == NULL)
		return -1;
	if(pheader)
		QNX4FS_PRINT(dformat);

	for (slot = 0; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++) {
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		if (qdp->i.status & QNX4FS_FILE_LINK) {
			namelen = QNX4FS_LONG_NAME_MAX;
			ino = link2ino(qdp);

			QNX4FS_PRINT("%9u %02X link -> (block %9u, index %1u) ",
			(unsigned)ino,
			(unsigned)(qdp->l.status),
			(unsigned)le32_to_cpu(qdp->l.block),
			(unsigned)qdp->l.index);

		} else if (qdp->i.status & QNX4FS_FILE_USED) {
				namelen = QNX4FS_SHORT_NAME_MAX;
				ino = direct2ino(block, slot);

			QNX4FS_PRINT("%9u %02X %9u %3u %9u %7u    ",
			(unsigned)ino,
			(unsigned)qdp->i.status,
			(unsigned)le32_to_cpu(qdp->i.size),
			(unsigned)le16_to_cpu(qdp->i.nxt),
			(unsigned)le32_to_cpu(qdp->i.xt0.start),
			(unsigned)le32_to_cpu(qdp->i.xt0.size));

		} else {
				if(qdp->i.status) 
				QNX4FS_PRINT("?  status %2X\n", (unsigned)(qdp->i.status));
			else
				QNX4FS_PRINT("~\n"); 
			continue;
		}

		namelen = strnlen(qdp->i.name, namelen);

		strncpy(dirnt->d_name, qdp->i.name, namelen); 
		dirnt->d_name[namelen] = 0;
		if(S_ISDIR(le16_to_cpu(qdp->i.mode)))
			QNX4FS_PRINT("/%s\n", dirnt->d_name);
		else
			QNX4FS_PRINT(" %s\n", dirnt->d_name);
	}
	brelse(bp);
	return 0;
}

int qnx4fs_lsdir (struct inode *dip)
{
	unsigned long phys;
	long nblocks, block;

	nblocks = qnx4fs_nblocks(dip);

	QNX4FS_PRINT(dformat);
	for (block = 0; block < nblocks; block++){
		phys = qnx4fs_map_block(dip, block);
		if (!phys)
			continue;
		qnx4fs_lsblock (dip->i_sb, phys, 0);
	}
	return 0;
}

int qnx4fs_find_block (struct super_block* sb, unsigned long block, char *name)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	int slot, namelen, res = 0;
	ino_t ino = 0;

	namelen = strlen(name);

	bp = sb_bread (sb, block - 1);
	if (bp == NULL)	return -1;

	for (slot = 0; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++) {

		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		if (qdp->i.status & QNX4FS_FILE_LINK) {
			ino = link2ino(qdp);
			if(memcmp(qdp->i.name, name, QNX4FS_LONG_NAME_MAX) != 0)
				continue;
		} else if (qdp->i.status & QNX4FS_FILE_USED) {
			ino = direct2ino(block, slot);
			if(namelen > QNX4FS_SHORT_NAME_MAX)
				continue;
			if(memcmp(qdp->i.name, name, QNX4FS_SHORT_NAME_MAX) != 0)
				continue;
		} else continue;
	
		printf(" Found entry <%s> block %lu, slot %d, ino %u",
			name, block, slot, (unsigned) ino);
		if (qdp->i.status & QNX4FS_FILE_USED) 
			printf(", size %u", (unsigned)qdp->i.size);
		printf("\n");
		res++;
	}

	brelse(bp);
	return res;
}


int qnx4fs_readdir (struct inode *dip, struct qnx4fs_dir *dirnt)
{
	struct buffer_head *bp;
	qnx4fs_dirent_t *qdp;
	unsigned long phys;
	long nblocks, block;
	int slot, namelen;
	ino_t ino;

	nblocks = qnx4fs_nblocks(dip);
	block = dirnt->d_count >> QNX4FS_DIRENTS_PER_BLOCK_BITS;
	slot  = dirnt->d_count & (QNX4FS_DIRENTS_PER_BLOCK-1);

	for (; block < nblocks; block++, slot = 0) {
		phys = qnx4fs_map_block(dip, block);
		if (!phys)
			continue;
		bp = sb_bread (dip->i_sb, phys - 1);
		if (bp == NULL)
			continue;
		qdp = (qnx4fs_dirent_t *)bp->b_data + slot;
		for (; slot < QNX4FS_DIRENTS_PER_BLOCK; slot++, qdp++) {
			if (!qdp->i.status)
				continue;
			if (qdp->i.status & QNX4FS_FILE_LINK) {
				namelen = QNX4FS_LONG_NAME_MAX;
				ino = link2ino(qdp);
			} else if (qdp->i.status & QNX4FS_FILE_USED) {
				namelen = QNX4FS_SHORT_NAME_MAX;
				ino = direct2ino(phys, slot);
			} else if (dip->i_ino != QNX4FS_ROOT_INO) {
				qdp->i.name[0] = '?';
				namelen = QNX4FS_SHORT_NAME_MAX;
				ino = direct2ino(phys, slot);
			} else {
				continue;
			}

			namelen = strnlen(qdp->i.name, namelen);
			strncpy(dirnt->d_name, qdp->i.name, namelen); 
			dirnt->d_name[namelen] = 0;
			dirnt->d_count = (block << QNX4FS_DIRENTS_PER_BLOCK_BITS) | slot;
			dirnt->d_namelen = namelen;
			dirnt->d_ino = ino;
			brelse(bp);
			return 1;
		}
		brelse(bp);
	}

	dirnt->d_count = (block << QNX4FS_DIRENTS_PER_BLOCK_BITS) | slot;
	return 0;
}

