/*
 * chkfsed.c               Copyright (C) 2004-2009  Dmitry Yakovlev 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

static const char help[] =
"\n\n        The hexadecimal QNX filesystem editor\n"
"\n"
" There are block, directory, inode and extent-blocks edit modes.\n"
"\n"
" Commands:\n"
"\n"
" n - next, p - prev, h - help, q - quit \n"
" i - edit current inode  (directory edit mode only)  \n"
" e - edit current file   (inode & directory edit mode) \n"
" l - list/edit directory (inode & directory edit mode) \n"
" x - edit extent-blocks  (inode edit mode only) \n"
" r - return to previous mode \n"
" u - undo last changes \n"
"\n"
" Write commands:\n"
"\n"
" ws offset string - write string\n"
" w  offset data   - write byte\n"
" ww offset data   - write 2 byte word\n"
" wl offset data   - write 4 byte word\n"
"\n"
" The offset is hexadecimal. The data is decimal, but may be expressed as \n"
" a hexadecimal or octal value by preceding it with leading '0x' or '0'. \n"
"\n"
" Press enter..."
;

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>
#include "qnx4fs_fs.h"

#define CMD_DONOTHING	1
#define CMD_NEXT	2
#define CMD_EXIT	4
#define CMD_PREV 	8
#define CMD_MODIFY	0x10
#define CMD_UNDO	0x20
#define CMD_ENTER	0x40
#define CMD_RETURN	0x80
#define CMD_INODE	0x100
#define CMD_XBLK	0x200
#define CMD_EDIT        0x800
 
#define UNDO_STRING      0x1000000
#define UNDO_DATA        0x100000

#define OFFSET_OF(X,Y) (int)(&((X*)0)->Y)

void chkfs_editor_help(){
	printf(help);
	fflush(stdout);
	getchar();
	printf("\n\n");
}

static int edit_buf(void* buf, int sze){
	static	unsigned offset, data, undo, udata, udatasize;
	static  char command[100];
	static  char cmdbuff[100];
	static  char undobuf[100];
	static  char first_call = 1;
	char *spt = cmdbuff;
	unsigned char* bf = buf;
	int i;
	
	if(first_call){
		printf(" The hexadecimal QNX4FS editor, press \"h\" for help... \n");
		first_call = 0;
	}
	printf(" : ");
	fflush(stdout);
	*cmdbuff = 0;

	fgets(cmdbuff, sizeof(cmdbuff), stdin);
	while(*spt == ' ') spt++;
	if(*spt == 0) return CMD_DONOTHING;
	switch(*spt){
	case 'n': undo = 0; return CMD_NEXT;
	case 'p': undo = 0; return CMD_PREV;
	case 'i': undo = 0; return CMD_INODE;
	case 'r': undo = 0; return CMD_EXIT;
	case 'x': undo = 0; return CMD_XBLK;
	case 'e': undo = 0; return CMD_EDIT;
	case 'q': exit(EXIT_SUCCESS);
	case 'h':
		chkfs_editor_help();
		return CMD_DONOTHING;
	case 'u': 
		if(!undo) {
			printf(" Nothing to undo\n");
			return CMD_DONOTHING;
		}
		if(undo == UNDO_STRING){
			memcpy(bf + offset, undobuf, udatasize);
			undo = 0;
		}
		if(undo == UNDO_DATA){
			memcpy(bf + offset, &udata, udatasize);
			undo = 0;
		}
		return CMD_UNDO;
	case 'l': undo = 0; return CMD_ENTER;
	case 'w': 
		i = 1;
		switch(spt[1]){
		case 's':
			if((i = sscanf(spt, " %*s %x%*c%[^\n]", &offset, command)) != 2) 
				goto quit;
			i = strlen(command);
			printf(" write string, len %d offset 0x%x <%s>\n", i, offset, command);
			if(offset + i > sze) {
				printf(" string <%s> is too long\n", command);
				return CMD_DONOTHING;
			}
			memcpy(undobuf, bf + offset, i);
			memcpy(bf + offset, command, i);
			undo = UNDO_STRING;
			udatasize = i;
			return CMD_MODIFY;
		case 'w': i = 2; break;
		case 'l': i = 4; break;
		}

		if(sscanf(cmdbuff, " %*s %x %i ", &offset, (int*)&data) == 2){
			if(offset + i > sze) {
				printf(" offset + sizeof(data) out of range\n");
				return CMD_DONOTHING;
			}
			memcpy(&udata, bf + offset, i);
			data = cpu_to_le32(data);
			memcpy(bf + offset, &data, i);
			undo = UNDO_DATA;
			udatasize = i;
			return CMD_MODIFY;
		}
	}
quit:
	printf("?\n");
	return CMD_DONOTHING;
}

static char fnamebuff[10000];

char* chkfs_add_name(char chr, char *str, int maxsize){
	int i = strlen(fnamebuff);
	char *stp = fnamebuff + i;

	if((i + maxsize + 2) > sizeof(fnamebuff))
		return stp; 

	if(maxsize <= 0) return stp;
	if(!str)         return stp;

	if((i == 1) && (*fnamebuff == '/'));
	else {
		if(chr)fnamebuff[i++] = chr;
	}
	while(maxsize-- > 0){
		if(*str == 0) break;
		if(*str < ' ') fnamebuff[i] = '_';
		else  fnamebuff[i] = *str;
		str++;
		i++;
	}
	fnamebuff[i] = 0;
	return stp;
}

char* chkfs_set_name(char *str, int maxsize){
	*fnamebuff = 0;
	return chkfs_add_name(0, str, maxsize);
}

char* chkfs_get_name(){
	return fnamebuff;
}

void chkfs_print_extents(qnx4fs_xt_t *xt, int offset, int size){
	int i = 0;

	for (i = 0; i < size; i++, xt++) {

		if((i & 1) == 0) printf("  %02x: ", offset + i * sizeof(*xt));
		printf("%9u (0x%08x), %6u; ", 
		(unsigned)le32_to_cpu(xt->start),
		(unsigned)le32_to_cpu(xt->start),
		(unsigned)le32_to_cpu(xt->size));

		if(i & 1) printf("\n");
	}
}

/* 
void chkfs_print_xblockp (qnx4fs_xb_t *xb, int part);

print contents of xblk
xblk = 512 bytes

Xblock output consists of 4 parts.
each xblock contains 60 extents

0 part = 
		u32		xb_next;
		u32		xb_prev;
		u8		xb_nxt;
		s32		xb_nblocks;
		extents[14]

1 part = extents[16]
2 part = extents[16]
3 part =
	extents[14]
	char		xb_signature[8];

if xblk = 0, 
	qnx4fs_xt_t	xb_xt0;

in other case xb_xt0 = 0;

*/

void chkfs_print_xblockp (qnx4fs_xb_t *xb, int part) {

	switch(part){
	case 0:
		printf(" next         0x%02x: \t%9u \t0x%08x \n", OFFSET_OF(qnx4fs_xb_t, xb_next), 
		(unsigned)le32_to_cpu(xb->xb_next), (unsigned)le32_to_cpu(xb->xb_next));

		printf(" prev         0x%02x: \t%9u \t0x%08x \n", OFFSET_OF(qnx4fs_xb_t, xb_prev), 
		(unsigned)le32_to_cpu(xb->xb_prev), (unsigned)le32_to_cpu(xb->xb_prev));
		
		printf(" nextents     0x%02x: \t%9u \t0x%02x \n", OFFSET_OF(qnx4fs_xb_t, xb_nxt), 
		(unsigned)xb->xb_nxt, xb->xb_nxt);

		printf(" nblocks      0x%02x: \t%9u  \t0x%08x \n", OFFSET_OF(qnx4fs_xb_t, xb_nblocks), 
		(unsigned)le32_to_cpu(xb->xb_nblocks), (unsigned)le32_to_cpu(xb->xb_nblocks));

		chkfs_print_extents(xb->xb_xts, 0x10, 14);
		break;
		
	case 1:
		chkfs_print_extents(xb->xb_xts + 14, 0, 16);
		break;
    
	case 2:
	
		chkfs_print_extents(xb->xb_xts + 30, 0, 16);
		break;

	case 3:
		chkfs_print_extents(xb->xb_xts + 46, 0, 14);
		
		printf("\n extent0    0x78: start: %9u (0x%08x) \tsize: %8u \n\n",
		(unsigned)le32_to_cpu(xb->xb_xt0.start), (unsigned)le32_to_cpu(xb->xb_xt0.start),
		(unsigned)le32_to_cpu(xb->xb_xt0.size));

		break;
		
	default:
		printf("\nchkfs: error in xblk part number\n");
		break;
	}
}

void chkfs_edit_xblk (struct inode *ip){
	struct buffer_head *bh;
	qnx4fs_xb_t *xb;
	unsigned xbnr;
	int xblkcnt = 0;
	int part = 0;
	
	xbnr = QNX4FS_I(ip)->i_xbnr0;
	if(!xbnr) {
		printf(" No extent-blocks defined\n");
		return;
	}
	for(;;){

		bh = sb_bread(ip->i_sb, xbnr - 1);

		if (bh == NULL){
			fprintf(stderr," Wrong xblk %u\n", xbnr);
			return;
		}
		xb = (qnx4fs_xb_t *)bh->b_data;
		
		printf("\n xblk edit mode: xblk %d, block %u, part %d\n", xblkcnt, xbnr, part);

		chkfs_print_xblockp(xb, part);
		chkfs_dump_area((char*)xb + part * 128, 128);

		switch(edit_buf((char*)xb + part * 128, 128)){

		case CMD_NEXT:
			if(part >= 3){
				if (xb->xb_next == 0) break;
				xbnr = le32_to_cpu(xb->xb_next);
				part = 0;
				xblkcnt++;
				break;
			}
			part++;
			break;
		case CMD_PREV:
			if(part <= 0){
				if (xb->xb_prev == 0) break;
				xbnr = le32_to_cpu(xb->xb_prev);
				part = 3;
				xblkcnt--;
				break;
			}
			part--;
			break;
		case CMD_EXIT:
			goto quitb;
		case CMD_UNDO:
		case CMD_MODIFY:
			if(sb_bwrite(ip->i_sb, bh) < 0) perror("write");
			break;

		}
		brelse(bh);
	}
quitb:
	brelse(bh);
}

char* chkfs_mode2str (unsigned mode){
	static char str[12] = {0};
	char *stp = str;
	int i;

	*stp++ = (S_ISDIR(mode))? 'd': ((S_ISLNK(mode))? 'l': '-');
	for(i = 0; i < 3; i++){
		*stp++ = (mode & S_IRUSR)? 'r': '-';
		*stp++ = (mode & S_IWUSR)? 'w': '-';
		*stp++ = (mode & S_IXUSR)? 'x': '-';
		mode <<= 3;
	}
	return str;
}

int chkfs_print_inode (struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_dinode_t *qip;
	unsigned long ptime;

	bp = sb_bread(ip->i_sb, ino2phys(ip->i_ino));
	if (bp == NULL)
		return -1;
	qip = (qnx4fs_dinode_t *)bp->b_data + ino2slot(ip->i_ino);

	printf(" size         0x%02x: \t%10u \t0x%08x \n", OFFSET_OF(qnx4fs_dinode_t, i.size), 
	(unsigned)le32_to_cpu(qip->i.size), (unsigned)le32_to_cpu(qip->i.size));

	printf(" xtnt0        0x%02x: \t%10u \t0x%08x \n", OFFSET_OF(qnx4fs_dinode_t, i.xt0.start),
	(unsigned)le32_to_cpu(qip->i.xt0.start), (unsigned)le32_to_cpu(qip->i.xt0.start));

	printf(" xtnt0 size   0x%02x: \t%10u \t0x%08x \n", OFFSET_OF(qnx4fs_dinode_t, i.xt0.size),
	(unsigned)le32_to_cpu(qip->i.xt0.size), (unsigned)le32_to_cpu(qip->i.xt0.size));

	printf(" xblk         0x%02x: \t%10u \t0x%08x \n", OFFSET_OF(qnx4fs_dinode_t, i.xb),
	(unsigned)le32_to_cpu(qip->i.xb), (unsigned)le32_to_cpu(qip->i.xb));
	
	ptime = le32_to_cpu(qip->i.ftime);
	printf(" ftime        0x%02x: \t0x%08x \t%s", OFFSET_OF(qnx4fs_dinode_t, i.ftime),
	(unsigned)le32_to_cpu(qip->i.ftime), ctime((time_t*)&ptime));

	ptime = le32_to_cpu(qip->i.mtime);
	printf(" mtime        0x%02x: \t0x%08x \t%s", OFFSET_OF(qnx4fs_dinode_t, i.mtime),
	(unsigned)le32_to_cpu(qip->i.mtime), ctime((time_t*)&ptime));

	ptime = le32_to_cpu(qip->i.atime);
	printf(" atime        0x%02x: \t0x%08x \t%s", OFFSET_OF(qnx4fs_dinode_t, i.atime),
	(unsigned)le32_to_cpu(qip->i.atime), ctime((time_t*)&ptime));

	ptime = le32_to_cpu(qip->i.ctime);
	printf(" ctime        0x%02x: \t0x%08x \t%s", OFFSET_OF(qnx4fs_dinode_t, i.ctime),
	(unsigned)le32_to_cpu(qip->i.ctime), ctime((time_t*)&ptime));
	
	printf(" nextents     0x%02x: \t%10u \t0x%04x \n", OFFSET_OF(qnx4fs_dinode_t, i.nxt),
	(unsigned)le16_to_cpu(qip->i.nxt), (unsigned)le16_to_cpu(qip->i.nxt));

	printf(" mode         0x%02x: \t0x%04x \t0%04o\t%s\n", OFFSET_OF(qnx4fs_dinode_t, i.mode), 
	(unsigned)le16_to_cpu(qip->i.mode), (unsigned)le16_to_cpu(qip->i.mode), 
	chkfs_mode2str(le16_to_cpu(qip->i.mode)));

	printf(" uid          0x%02x: \t0x%04x \n", OFFSET_OF(qnx4fs_dinode_t, i.uid),
	(unsigned)le16_to_cpu(qip->i.uid));

	printf(" gid          0x%02x: \t0x%04x \n", OFFSET_OF(qnx4fs_dinode_t, i.gid),
	(unsigned)le16_to_cpu(qip->i.gid));

	printf(" nlink        0x%02x:  \t%10u \t0x%04x \n", OFFSET_OF(qnx4fs_dinode_t, i.nlink),
	(unsigned)le16_to_cpu(qip->i.nlink), (unsigned)le16_to_cpu(qip->i.nlink));

	printf(" type         0x%02x: \t0x%02x \n", OFFSET_OF(qnx4fs_dinode_t, i.type), qip->i.type);

	printf(" status       0x%02x: \t0x%02x ", OFFSET_OF(qnx4fs_dinode_t, i.status), qip->i.status);

	if(qip->i.status & QNX4FS_FILE_USED)     printf("USED ");
	if(qip->i.status & QNX4FS_FILE_MODIFIED) printf("MODIFIED ");
	if(qip->i.status & QNX4FS_FILE_BUSY)     printf("BUSY ");
	if(qip->i.status & QNX4FS_FILE_LINK)     printf("LINK ");
	if(qip->i.status & QNX4FS_FILE_INODE)    printf("INODE ");
	if(qip->i.status & QNX4FS_FSYS_CLEAN)    printf("FSYS_CLEAN ");
	if(qip->i.status & QNX4FS_FILE_GROWN)    printf("GROWN");
	printf("\n\n");

	brelse(bp);
	return 0;
}

void chkfs_edit_inode (struct inode *ip){
	struct buffer_head *bh;
	qnx4fs_dinode_t *qip;
	
	for(;;){
		printf("\n inode edit mode, ino %u, block %u, slot %u\n", 
		(unsigned)ip->i_ino, (unsigned)ino2phys(ip->i_ino), 
		(unsigned)ino2slot(ip->i_ino));

		chkfs_print_inode(ip);
		bh = sb_bread(ip->i_sb, ino2phys(ip->i_ino));
		if (bh == NULL) {
			printf(" Wrong inode %u\n", ip->i_ino);
			return;
		}
		qip = (qnx4fs_dinode_t *)bh->b_data + ino2slot(ip->i_ino);
		printf("<%s>\n", chkfs_get_name());
		chkfs_dump_area(qip, 64);

		switch(edit_buf(qip, 64)){

		case CMD_XBLK:
			if(qnx4fs_read_inode(ip) < 0){
				printf(" Can't read inode %d\n", ip->i_ino);
				break;
			}
			chkfs_edit_xblk(ip);
			break;
		case CMD_EDIT:
			if(qnx4fs_read_inode(ip) < 0){
				printf(" Can't read inode %d\n", ip->i_ino);
				break;
			}
			chkfs_edit_file(ip);
			break;
		case CMD_ENTER:
			if(qnx4fs_read_inode(ip) < 0){
				printf(" Can't read inode %d\n", ip->i_ino);
				break;
			}
			chkfs_edit_dir(ip);
			break;
		case CMD_EXIT:
			goto quitb;
		case CMD_UNDO:	// fall
		case CMD_MODIFY:
			if(sb_bwrite(ip->i_sb, bh) < 0) perror("write");
			break;

		}
		brelse(bh);
	}
quitb:
	brelse(bh);
}

void chkfs_edit_block(struct super_block* sb, unsigned long block){
	unsigned block_offset = 0;
	struct buffer_head *bh;

	for(;;){
		bh = sb_bread(sb, block - 1);

		if(!bh)
			goto quit;
		printf("\n block edit mode, block %lu, offset 0x%x\n", block, block_offset);
		chkfs_dump_area(bh->b_data + block_offset, 256);

		switch(edit_buf(bh->b_data + block_offset, 256)){

		case CMD_NEXT:
			if(block_offset == 0){
				block_offset = 256;
				break;
			}
			block_offset = 0;
			block++;
			break;

		case CMD_PREV:
			if(block_offset == 256){
				block_offset = 0;
				break;
			}
			block_offset = 256;
			if(block <= 1) break;
			block--;
			break;

		case CMD_EXIT:
			goto quitb;
			break;

		case CMD_UNDO:	// fall
		case CMD_MODIFY:
			if(sb_bwrite(sb, bh) < 0) perror("write");
			break;

		}
		brelse(bh);
	}
quitb:
	brelse(bh);
quit:	;

}

void chkfs_edit_dir (struct inode *dip) {
	struct buffer_head *bh;
	struct inode dinode;
	qnx4fs_dinode_t *qip;
	unsigned long phys;
	long nblocks, block = 0;
	int slot = 0;
	char *str = 0;



	nblocks = qnx4fs_nblocks(dip);

	for(;;){
		phys = qnx4fs_map_block(dip, block);

		if(!phys) {
			printf(" Can't map block %ld\n", block);
			goto quit;
		}

		printf("\n dir edit mode, ino %u, block %lu, slot %d\n",
		dip->i_ino, phys, slot);

		bh = sb_bread(dip->i_sb, phys - 1);
		if(!bh) {
			printf(" Can't read block %ld, phys %lu\n", block, phys);
			goto quit;
		}

		str = chkfs_add_name('/', (char*)((qnx4fs_dirent_t *)bh->b_data + slot),
		QNX4FS_LONG_NAME_MAX);

		qnx4fs_lsblock (dip->i_sb, phys, 1);

		printf("<%s>\n", chkfs_get_name());
		chkfs_dump_area((qnx4fs_dirent_t *)bh->b_data + slot, 64);

		switch(edit_buf((qnx4fs_dirent_t *)bh->b_data + slot, 64)){

		case CMD_NEXT:
			if((slot + 1) >= 8){
				if((block + 1)>= nblocks) break;
				slot = 0;
				block++;
				break;
			}
			slot++;
			break;

		case CMD_PREV:
			if((slot - 1) < 0){
				if((block - 1) < 0) break;
				slot = 7;
				block--;
				break;
			}
			slot--;
			break;

		case CMD_ENTER:
			qip = (qnx4fs_dinode_t *)bh->b_data + slot;
			if (qip->i.status & QNX4FS_FILE_LINK) {
				dinode.i_ino = link2ino(qip);
			}
			else {
				dinode.i_ino = direct2ino(phys, slot);
			}
			dinode.i_sb = dip->i_sb;

			if(qnx4fs_read_inode(&dinode) < 0){
				printf(" Cant read inode %u\n", (unsigned)dinode.i_ino);
			}
			else {
				chkfs_edit_dir(&dinode);
			}
			break;

		case CMD_INODE:
			qip = (qnx4fs_dinode_t *)bh->b_data + slot;
			if (qip->i.status & QNX4FS_FILE_LINK) {
				dinode.i_ino = link2ino(qip);
			}
			else {
				dinode.i_ino = direct2ino(phys, slot);
			}
			dinode.i_sb = dip->i_sb;
			chkfs_edit_inode(&dinode);
			break;

		case CMD_EDIT:
			qip = (qnx4fs_dinode_t *)bh->b_data + slot;
			if (qip->i.status & QNX4FS_FILE_LINK) {
				dinode.i_ino = link2ino(qip);
			}
			else {
				dinode.i_ino = direct2ino(phys, slot);
			}
			dinode.i_sb = dip->i_sb;
			chkfs_edit_file(&dinode);
			break;
		case CMD_EXIT:
			goto quitb;
			break;

		case CMD_UNDO: // fall
		case CMD_MODIFY:
			if(sb_bwrite(dip->i_sb, bh) < 0) perror("write");
			break;

		}
		brelse(bh);
		chkfs_restore_name(str);
	}
quitb:
	brelse(bh);
quit:	
	chkfs_restore_name(str);
}

void chkfs_edit_file (struct inode *ip) {
	struct buffer_head *bh;
	unsigned block_offset = 0;
	unsigned long phys;
	long nblocks, block = 0;


	if(qnx4fs_read_inode(ip)){
			printf(" Can't read inode %u (%u,%u)\n", 
			(unsigned)ip->i_ino, (unsigned)ino2phys(ip->i_ino), 
			(unsigned)ino2slot(ip->i_ino));
			goto quit;
	}
	nblocks = qnx4fs_nblocks(ip);

	for(;;){
		phys = qnx4fs_map_block(ip, block);

		if(!phys) {
			printf(" Can't map block %ld\n", block);
			goto quit;
		}

		printf("\n file edit mode, block %u of %u, phys %u, offset %u\n",
		(unsigned)block + 1, (unsigned)nblocks, (unsigned)phys, block_offset);

		bh = sb_bread(ip->i_sb, phys - 1);
		if(!bh) {
			printf(" Can't read block %ld, phys %lu\n", block, phys);
			goto quit;
		}

		printf("<%s>\n", chkfs_get_name());
		chkfs_dump_area(bh->b_data + block_offset, 256);

		switch(edit_buf(bh->b_data + block_offset, 256)){

		case CMD_NEXT:
			if(block_offset == 0){
				block_offset = 256;
				break;
			}
			if(block + 1 >= nblocks) break;
			block_offset = 0;
			block++;
			break;

		case CMD_PREV:
			if(block_offset == 256){
				block_offset = 0;
				break;
			}
			if(block < 1) break;
			block_offset = 256;
			block--;
			break;

		case CMD_EXIT:
			goto quitb;
			break;

		case CMD_UNDO:	// fall
		case CMD_MODIFY:
			if(sb_bwrite(ip->i_sb, bh) < 0) perror("write");
			break;

		}

		brelse(bh);
	} // for(;;)
quitb:
	brelse(bh);
quit:	
	;
}

void chkfs_edit_dir_blk (struct super_block* sb, unsigned long block) {
	struct buffer_head *bh;
	struct inode dinode;
	qnx4fs_dinode_t *qip;
	int slot = 0;

	for(;;){
		printf("\n block dir edit mode, ino %u, block %lu, slot %d\n",
		(unsigned)direct2ino(block, slot), block, slot);
		qnx4fs_lsblock (sb, block, 1);

		bh = sb_bread(sb, block - 1);
		if(!bh) {
			printf(" Can't read block %lu\n", block);
			goto quit;
		}

		chkfs_dump_area((qnx4fs_dirent_t *)bh->b_data + slot, 64);

		switch(edit_buf((qnx4fs_dirent_t *)bh->b_data + slot, 64)){

		case CMD_NEXT:
			if((slot + 1) >= 8){
				slot = 0;
				block++;
				break;
			}
			slot++;
			break;

		case CMD_PREV:
			if((slot - 1) < 0){
				if(block == 0) break;
				slot = 7;
				block--;
				break;
			}
			slot--;
			break;

		case CMD_ENTER:
			qip = (qnx4fs_dinode_t *)bh->b_data + slot;
			if (qip->i.status & QNX4FS_FILE_LINK) {
				dinode.i_ino = link2ino(qip);
			}
			else {
				dinode.i_ino = direct2ino(block, slot);
			}
			dinode.i_sb = sb;

			if(qnx4fs_read_inode(&dinode) < 0){
				printf(" Cant read inode %u\n", (unsigned)dinode.i_ino);
			}
			else {
				chkfs_edit_dir(&dinode);
			}
			break;

		case CMD_EDIT:
			qip = (qnx4fs_dinode_t *)bh->b_data + slot;
			if (qip->i.status & QNX4FS_FILE_LINK) {
				dinode.i_ino = link2ino(qip);
			}
			else {
				dinode.i_ino = direct2ino(block, slot);
			}
			dinode.i_sb = sb;
			chkfs_edit_file(&dinode);
			break;

		case CMD_INODE:
			qip = (qnx4fs_dinode_t *)bh->b_data + slot;
			if (qip->i.status & QNX4FS_FILE_LINK) {
				dinode.i_ino = link2ino(qip);
			}
			else {
				dinode.i_ino = direct2ino(block, slot);
			}
			dinode.i_sb = sb;
			chkfs_edit_inode(&dinode);
			break;

		case CMD_EXIT:
			goto quitb;
			break;

		case CMD_UNDO: // fall
		case CMD_MODIFY:
			if(sb_bwrite(sb, bh) < 0) perror("write");
			break;

		}
		brelse(bh);
	}
quitb:
	brelse(bh);
quit:	;

}
