/*
 * file.c                Copyright (C) 2004-2009 Dmitry Yakovlev
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include "qnx4fs_fs.h"

int qnx4fs_read_file (struct inode *ip, int fd)
{
	unsigned long phys;
	long nblocks, block, sze;
	struct buffer_head *bp;

	nblocks = qnx4fs_nblocks(ip);
	sze = ip->i_size;
	for (block = 0; block < nblocks; block++){

		phys = qnx4fs_map_block(ip, block);
		if (!phys) return -1;

		bp = sb_bread(ip->i_sb, phys - 1);
		if (bp == NULL){
			return -1;
		}

		if(write(fd, bp->b_data, (sze > QNX4FS_BLOCK_SIZE)?
		QNX4FS_BLOCK_SIZE : sze) < 0){
			brelse(bp);
			return -1;
		}
		brelse(bp);
		sze -= QNX4FS_BLOCK_SIZE;
	}
	return 0;
}

void qnx4fs_print_xblocks (struct inode *ip)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	qnx4fs_xt_t *xt;
	long xbnr, offset;
	int nxt, i;

	xbnr = QNX4FS_I(ip)->i_xbnr0;
	offset = QNX4FS_I(ip)->i_xt0.size;
	while (xbnr) {
		bp = sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL){
			fprintf(stderr," wrong xblk xbnr %ld\n", xbnr);
			return;
		}
		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			fprintf(stderr," wrong xblk signature %ld\n", xbnr);
			brelse(bp);
			return;
		}

		QNX4FS_PRINT(
		"   xblk        next      prev    nxtnts   nblocks     fxtnt   size\n"
		"%9u %9u %9u %9u %9u %9u %6u\n",
		(unsigned)xbnr,
		(unsigned)le32_to_cpu(xb->xb_next),
		(unsigned)le32_to_cpu(xb->xb_prev),
		(unsigned)xb->xb_nxt,
		(unsigned)le32_to_cpu(xb->xb_nblocks),
		(unsigned)le32_to_cpu(xb->xb_xt0.start),
		(unsigned)le32_to_cpu(xb->xb_xt0.size));

		nxt = xb->xb_nxt;
		xt = xb->xb_xts;
		offset = 0;
		QNX4FS_PRINT("  extent       size    extent       size \n");
		for (i = 0; i < nxt; i++) {
			QNX4FS_PRINT("%9u %9u ", 
			(unsigned)le32_to_cpu(xt->start),
			(unsigned)le32_to_cpu(xt->size));
			if(i & 1) 
				QNX4FS_PRINT("\n");
			offset += le32_to_cpu(xt->size);
			xt++;
		}
		QNX4FS_PRINT("\n %u blocks in %u extents %s\n", 
		(unsigned)offset, (unsigned)xb->xb_nxt, 
		((le32_to_cpu(xb->xb_nblocks) == offset) ? "OK" : "ERROR"));

		xbnr = le32_to_cpu(xb->xb_next);
		brelse (bp);
	}
}

int qnx4fs_check_xblk (struct inode *ip, qnx4fs_xb_t *xblk, unsigned block){
	unsigned nxt, blocks;
	struct super_block* sb = ip->i_sb;
	qnx4fs_xt_t *xt;
	int res = 0;
	
	if (memcmp(xblk->xb_signature, "IamXblk\0", 8)) {
		fprintf(stderr,"\n <%s>\n xblk %u wrong signature \n", 
		chkfs_get_name(), block);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	nxt = xblk->xb_nxt;
	
	if(nxt > QNX4FS_MAX_XTS_PER_XBLK){
		fprintf(stderr,"\n <%s>\n xblk %u wrong extents number %d (max = %d)\n", 
		chkfs_get_name(), block, nxt, QNX4FS_MAX_XTS_PER_XBLK);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);

		return -1;
	}
	
	xt = xblk->xb_xts;

	for (blocks = 0; nxt > 0; nxt--) {

		if(qnx4fs_check_range(sb, le32_to_cpu(xt->start), le32_to_cpu(xt->size)) != 0){
			fprintf(stderr, "\n <%s>\n xblk %u wrong extent %d, start %u, size %u\n", 
			chkfs_get_name(), block, xblk->xb_nxt - nxt, le32_to_cpu(xt->start), le32_to_cpu(xt->size));
			if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
				chkfs_edit_inode(ip);
			return -1;
		}

		res |= qnx4fs_bcheck(sb, le32_to_cpu(xt->start), le32_to_cpu(xt->size));
		res |= qnx4fs_btest_and_set(sb, le32_to_cpu(xt->start), le32_to_cpu(xt->size));

		/* size in blocks, count blocks number	*/
		blocks += le32_to_cpu(xt->size);
		xt++;
	}

	if(blocks != le32_to_cpu(xblk->xb_nblocks)){

		fprintf(stderr,"\n <%s>\n xblk %u wrong blocks number %u, should be %u\n", 
		chkfs_get_name(), block, (unsigned)le32_to_cpu(xblk->xb_nblocks), blocks );

		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);

		return -1;
	}
	return res;
}

int qnx4fs_check_zero_file (struct inode *ip) {

	if(QNX4FS_I(ip)->i_xt0.start != 0 || QNX4FS_I(ip)->i_xt0.size != 0){
		fprintf(stderr, "\n <%s>\n wrong extent 0 of zero length file: start %u, size %u\n", 
		chkfs_get_name(), QNX4FS_I(ip)->i_xt0.start, QNX4FS_I(ip)->i_xt0.size);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	if(QNX4FS_I(ip)->i_xbnr0){
		fprintf(stderr, "\n <%s>\n wrong xblk link of zero length file: %u\n", 
		chkfs_get_name(), (unsigned)QNX4FS_I(ip)->i_xbnr0);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}
	if(QNX4FS_I(ip)->i_nxt != 0){
		fprintf(stderr, "\n <%s>\n wrong extents number of zero length file: %u\n", 
		chkfs_get_name(), (unsigned)QNX4FS_I(ip)->i_nxt);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	if (S_ISDIR(ip->i_mode)){
		fprintf(stderr, "\n <%s>\n zero length directory\n", chkfs_get_name());
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}
	return 0;
}

int qnx4fs_check_file (struct inode *ip) {
	struct buffer_head *bp;
	
	qnx4fs_xb_t *xb;
	unsigned long xbnr, xbprev, blocks, nblocks, xtnts, nxtnts;
	int res;

	res = 0;
	xtnts = nxtnts = QNX4FS_I(ip)->i_nxt;
	blocks = QNX4FS_I(ip)->i_xt0.size;
	if(blocks) nxtnts--;
	nblocks = qnx4fs_nblocks(ip);

	if (!(QNX4FS_I(ip)->i_status & QNX4FS_FILE_USED)
	&& (ip->i_ino != QNX4FS_LOST_AND_FOUND_INO)) {
		fprintf(stderr,"\n <%s>\n bad inode status %04x ino %u, block %u, slot %u\n", 
		chkfs_get_name(), QNX4FS_I(ip)->i_status, (unsigned)ip->i_ino, (unsigned)ino2phys(ip->i_ino), 
		(unsigned)ino2slot(ip->i_ino));
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}
	
	if (!S_ISREG(ip->i_mode) && !S_ISDIR(ip->i_mode) && !S_ISLNK(ip->i_mode)	
	&& (ip->i_ino != QNX4FS_LOST_AND_FOUND_INO)) {
		fprintf(stderr,"\n <%s>\n bad inode mode %04x ino %u, block %u, slot %u\n", 
		chkfs_get_name(), ip->i_mode, (unsigned)ip->i_ino, (unsigned)ino2phys(ip->i_ino), 
		(unsigned)ino2slot(ip->i_ino));
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	if(nblocks == 0)
		return qnx4fs_check_zero_file (ip);

	if(qnx4fs_check_range(ip->i_sb, QNX4FS_I(ip)->i_xt0.start, QNX4FS_I(ip)->i_xt0.size) != 0){

		fprintf(stderr, "\n <%s>\n Wrong extent 0 range, start = %u, size = %u\n", 
		chkfs_get_name(), QNX4FS_I(ip)->i_xt0.start, QNX4FS_I(ip)->i_xt0.size);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}
	
	res |= qnx4fs_bcheck(ip->i_sb, QNX4FS_I(ip)->i_xt0.start, QNX4FS_I(ip)->i_xt0.size);
	res |= qnx4fs_btest_and_set(ip->i_sb, QNX4FS_I(ip)->i_xt0.start, QNX4FS_I(ip)->i_xt0.size);

	xbnr = QNX4FS_I(ip)->i_xbnr0;

	if (xbnr > QNX4FS_SB(ip->i_sb)->s_blocks){
		fprintf(stderr,"\n <%s>\n inode xblk link out of range %lu\n", 
		chkfs_get_name(), xbnr);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	xbprev = 0;

	while (xbnr) {

		res |= qnx4fs_bcheck(ip->i_sb, xbnr, 1);
		res |= qnx4fs_btest_and_set(ip->i_sb, xbnr, 1);

		bp = sb_bread(ip->i_sb, xbnr - 1);

		if (bp == NULL){
			fprintf(stderr," This should not happen!\n Wrong xblk xbnr %lu\n", xbnr);
			return -1;
		}
		xb = (qnx4fs_xb_t *)bp->b_data;

		if((res |= qnx4fs_check_xblk(ip, xb, xbnr)) == -1){
			brelse (bp);
			return -1;
		}

		if(xbprev != le32_to_cpu(xb->xb_prev)){
			fprintf(stderr,"\n <%s>\n xblk %lu wrong prev link: %u shuld be %lu\n",
			chkfs_get_name(), xbnr, (unsigned) le32_to_cpu(xb->xb_prev), xbprev);
			if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
				chkfs_edit_inode(ip);
			res = -1;
		}

		if (le32_to_cpu(xb->xb_next) > QNX4FS_SB(ip->i_sb)->s_blocks){
			fprintf(stderr,"\n <%s>\n xblk %lu next link out of range: %u\n", 
			chkfs_get_name(), xbnr, (unsigned)le32_to_cpu(xb->xb_next));
			if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
				chkfs_edit_inode(ip);
			brelse (bp);
			return -1;
		}

		if(xbprev == 0){

			if(QNX4FS_I(ip)->i_xt0.start != le32_to_cpu(xb->xb_xt0.start)){
				fprintf(stderr,"\n <%s>\n xblk %lu wrong first extent link: %u, should be %u\n",
				chkfs_get_name(), xbnr, (unsigned) le32_to_cpu(xb->xb_xt0.start), 
				(unsigned) QNX4FS_I(ip)->i_xt0.start);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
					chkfs_edit_inode(ip);
				res = -1;
			}

			if(QNX4FS_I(ip)->i_xt0.size != le32_to_cpu(xb->xb_xt0.size)){
				fprintf(stderr,"\n <%s>\n xblk %lu wrong first extent size: %u, should be %u\n",
				chkfs_get_name(), xbnr, (unsigned) le32_to_cpu(xb->xb_xt0.size), 
				(unsigned) QNX4FS_I(ip)->i_xt0.size);
				if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
					chkfs_edit_inode(ip);
				res = -1;
			}
		}

		nxtnts -= xb->xb_nxt;
		blocks += le32_to_cpu(xb->xb_nblocks);
		xbprev = xbnr;
		xbnr = le32_to_cpu(xb->xb_next);
		brelse (bp);
	}

	if(blocks != nblocks && !(QNX4FS_I(ip)->i_status & QNX4FS_FILE_GROWN)){
		fprintf(stderr,"\n <%s>\n wrong blocks number: %lu, should be %lu\n",
		chkfs_get_name(), blocks, nblocks);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	if(nxtnts != 0){
		fprintf(stderr,"\n <%s>\n wrong inode extents number: %lu, should be %lu\n", 
		chkfs_get_name(), xtnts, xtnts - nxtnts);
		if(chkfs_press_keym(chkfs_EDIT) == chkfs_EDIT)
			chkfs_edit_inode(ip);
		return -1;
	}

	return res;
}

#if 0
unsigned long qnx4fs_map_block (struct inode *ip, long block)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	qnx4fs_xt_t *xt;
	unsigned long phys;
	long xbnr, offset;
	int nxt;

	if (block < QNX4FS_I(ip)->i_xt0.size)
		return (QNX4FS_I(ip)->i_xt0.start + block);
	xbnr = QNX4FS_I(ip)->i_xbnr;
	offset = QNX4FS_I(ip)->i_xt0.size;
	while (xbnr) {
		bp = sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL)	return 0;

		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse(bp);
			return 0;
		}
		nxt = xb->xb_nxt;
		xt = xb->xb_xts;

		for (; nxt > 0; nxt--) {
			if (block < offset + xt->size) {
				phys = xt->start + block - offset;
				brelse(bp);
				return phys;
			}
			offset += xt->size;
			xt++;
		}
		xbnr = xb->xb_next;
		brelse (bp);
	}
	return 0;
}
#else
unsigned long
qnx4fs_map_block (struct inode *ip, long block)
{
	struct buffer_head *bp;
	qnx4fs_xb_t *xb;
	qnx4fs_xt_t *xt;
	unsigned long phys;
	long xbnr, offset;
	int xtnr, nxt;

	xbnr = QNX4FS_I(ip)->i_xbnr;
	xtnr = QNX4FS_I(ip)->i_xtnr;
	if (block >= QNX4FS_I(ip)->i_offset) {
		if (block < QNX4FS_I(ip)->i_offset_end) {
			phys = (QNX4FS_I(ip)->i_phys + block - QNX4FS_I(ip)->i_offset);
			goto found1;
		} else for (offset = QNX4FS_I(ip)->i_offset_end;
			    xbnr;
			    xbnr = le32_to_cpu(xb->xb_next),
			    brelse(bp),
			    xtnr = -1) {
			bp = sb_bread (ip->i_sb, xbnr - 1);
			if (bp == NULL){
				goto outerr;
			}
			xb = (qnx4fs_xb_t *)bp->b_data;
			if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
				brelse (bp);
				goto outerr;
			}
			nxt = xb->xb_nxt;
			if (xtnr >= nxt - 1)
				continue;
			xt = &xb->xb_xts[++xtnr];
			for (; xtnr < nxt; xtnr++, xt++) {
				offset += le32_to_cpu(xt->size);
				if (block < offset) {
					phys = le32_to_cpu(xt->start)
					       + le32_to_cpu(xt->size)
					       + block - offset;
					QNX4FS_I(ip)->i_offset = offset - le32_to_cpu(xt->size);
					QNX4FS_I(ip)->i_offset_end = offset;
					goto found;
				}
			}
		};
	} else if (block < QNX4FS_I(ip)->i_xt0.size) {
		phys = (QNX4FS_I(ip)->i_xt0.start + block);
		goto found1;
	} else for (offset = QNX4FS_I(ip)->i_offset;
		    xbnr;
		    xbnr = le32_to_cpu(xb->xb_prev),
		    brelse(bp),
		    xtnr = QNX4FS_MAX_XTS_PER_XBLK) {
		bp = sb_bread(ip->i_sb, xbnr - 1);
		if (bp == NULL){
			goto outerr;
		}
		xb = (qnx4fs_xb_t *)bp->b_data;
		if (memcmp(xb->xb_signature, "IamXblk\0", 8)) {
			brelse (bp);
			goto outerr;
		}
		if (xtnr <= 0)
			continue;
		nxt = xb->xb_nxt;
		xt = &xb->xb_xts[--xtnr];
		for (; xtnr >= 0; xtnr--, xt--) {
			offset -= le32_to_cpu(xt->size);
			if (block >= offset) {
				phys = le32_to_cpu(xt->start) + block - offset;
				QNX4FS_I(ip)->i_offset = offset;
				QNX4FS_I(ip)->i_offset_end = offset + le32_to_cpu(xt->size);
				goto found;
			}
		}
	};
outerr:
	return 0;
found:
	QNX4FS_I(ip)->i_phys = le32_to_cpu(xt->start);
	QNX4FS_I(ip)->i_xtnr = xtnr;
	QNX4FS_I(ip)->i_xbnr = xbnr;
	brelse(bp);
found1:
	return phys;
}
#endif
      
                     
                    
            
