#ifndef __QNX4FS_FS_H
#define __QNX4FS_FS_H

#include <fcntl.h>
#include <unistd.h>

#if (defined(__QNX__) && !defined(__QNXNTO__)) 
// || defined(WIN32) 

#define inline

#endif

#if defined(WIN32)

#define _S_IFLNK    0xA000
#define S_ISLNK(m)  (((m)&_S_IFMT)==_S_IFLNK)

#endif

#define QNX4FS_ROOT_INO			8
#define QNX4FS_INODES_INO		9
#define QNX4FS_BOOT_INO			10
#define QNX4FS_ALTBOOT_INO		11
#define QNX4FS_LOST_AND_FOUND_INO	12

#define QNX4FS_LINK_MAX			65535

#define QNX4FS_SUPER_MAGIC		0x34584e51 /* little endian "QNX4" */

#define QNX4FS_BLOCK_SIZE		512
#define QNX4FS_BLOCK_SIZE_BITS		9
#define QNX4FS_DIRENT_SIZE		64
#define QNX4FS_DIRENT_SIZE_BITS		6
#define QNX4FS_INODES_PER_BLOCK		8
#define QNX4FS_INODES_PER_BLOCK_BITS	3
#define QNX4FS_DIRENTS_PER_BLOCK	QNX4FS_INODES_PER_BLOCK
#define QNX4FS_DIRENTS_PER_BLOCK_BITS	QNX4FS_INODES_PER_BLOCK_BITS

#define QNX4FS_BOOT_BLOCK		1L
#define QNX4FS_ROOT_BLOCK		2L
#define QNX4FS_BITMAP_BLOCK		3L

#define QNX4FS_MAX_XTS_PER_XBLK		60
#define QNX4FS_LONG_NAME_MAX		48
#define QNX4FS_SHORT_NAME_MAX		16

#define QNX4FS_FILE_USED	0x01
#define QNX4FS_FILE_MODIFIED	0x02
#define QNX4FS_FILE_BUSY	0x04
#define QNX4FS_FILE_LINK	0x08
#define QNX4FS_FILE_INODE	0x10
#define QNX4FS_FSYS_CLEAN	0x20
#define QNX4FS_FILE_GROWN	0x40

#define QNX4FS_DIR_EXPAND_DELTA		4
#define QNX4FS_INODES_EXPAND_DELTA	16

typedef signed   char s8;
typedef unsigned char u8;

typedef signed   short s16;
typedef unsigned short u16;

typedef signed   int s32;
typedef unsigned int u32;

typedef struct qnx4fs_xt {
	u32 start;
	u32 size;
} qnx4fs_xt_t;

typedef struct qnx4fs_xb {
	u32		xb_next;
	u32		xb_prev;
	u8		xb_nxt;
	char		xb_spare[3];
	s32		xb_nblocks;
	qnx4fs_xt_t	xb_xts[QNX4FS_MAX_XTS_PER_XBLK];
	char		xb_signature[8];
	qnx4fs_xt_t	xb_xt0;
} qnx4fs_xb_t;

struct qnx4fs_dinode {
	char		name[QNX4FS_SHORT_NAME_MAX];
	u32		size;
	qnx4fs_xt_t	xt0;
	u32		xb;
	u32		ftime;
	u32		mtime;
	u32		atime;
	u32		ctime;
	u16		nxt;
	u16		mode;
	u16		uid;
	u16		gid;
	u16		nlink;
	char		zero[4];
	u8		type;
	u8		status;
};

struct qnx4fs_dlink {
	char	name[QNX4FS_LONG_NAME_MAX];
	u32	block;
	u8	index;
	char	spare[10];
	u8	status;
};

struct qnx4fs_dot_entry {
	char	signature[8];
	u32	block;
};

struct qnx4fs_dir {
	char	d_name[512];
	long	d_count;
	long    d_namelen;
	long    d_ino;
};

typedef union qnx4fs_dirent  {
	struct qnx4fs_dinode i;
	struct qnx4fs_dlink  l;
} qnx4fs_dirent_t;

typedef qnx4fs_dirent_t qnx4fs_dinode_t;

#define QNX4FS_WARN  printf 

#define QNX4FS_PRINT printf

/* 
 * qnx4fs filesystem in-core inode info all of these 
 * are in cpu byte order
 */
struct qnx4fs_inode_info {
	int		i_nxt;
	int		i_xtnr;
	qnx4fs_xt_t	i_xt0;
	unsigned long	i_xbnr;
	unsigned long	i_xbnr0;
	unsigned long	i_xbnr_last;
	long		i_offset;
	long		i_offset_end;
	unsigned long	i_phys;
	int		i_linked;
	unsigned	i_status;
};

/* qnx4fs in-core superblock data */
struct qnx4fs_sb_info {
	ino_t			s_bitmap_ino;
	unsigned long		s_bitmap_blocks;
	unsigned long		s_blocks;
	struct inode		*s_inodes;
	struct inode		*s_root_dir;
	unsigned char		*s_bitmap_new;
};

struct super_block {
	struct qnx4fs_sb_info s_fs_info;
	unsigned long		s_blocks_offset;
	int s_fd;
	char *s_devname;
};

struct inode {
	struct qnx4fs_inode_info i_info;
	struct super_block* i_sb;
	u32 i_size;
	u32 i_mode;
	u32 i_nlink;
	u32 i_blocks;
	u32 i_ino;
};

struct buffer_head {
	struct buffer_head *next;
	unsigned long       blknum;
	unsigned char b_data[QNX4FS_BLOCK_SIZE];
};

extern struct buffer_head* chkfs_sb_bread(struct super_block* sbp, u32 phys);
extern void   chkfs_brelse(struct buffer_head*);
extern int    chkfs_bwrite(struct super_block* sb, struct buffer_head* bh);
extern int    chkfs_press_keym (int mode);

#define sb_bread(X,Y) chkfs_sb_bread((X),(Y))
#define sb_bwrite(X,Y) chkfs_bwrite((X),(Y))
#define brelse(X)     chkfs_brelse((X))

#ifdef BIG_ENDIAN_CPU

#define le32_to_cpu(x) _swab32((u32)(x))
#define le16_to_cpu(x) _swab16((u16)(x))

#else

#define le32_to_cpu(x) (x)
#define le16_to_cpu(x) (x)

#define cpu_to_le32(x) (x)
#define cpu_to_le16(x) (x)

#endif

#define chkfs_EBIT_NSET        2
#define chkfs_EBIT_SET         8

/*
 * hweightN: returns the hamming weight (i.e. the number
 * of bits set) of a N-bit word
 */

static inline unsigned int generic_hweight8(unsigned int w) {
       unsigned int res = (w & 0x55) + ((w >> 1) & 0x55);
       res =  (res & 0x33) + ((res >> 2) & 0x33);
       return (res & 0x0F) + ((res >> 4) & 0x0F);
}

#ifdef BIG_ENDIAN_CPU

static inline u16 _swab16(u16 x) {
return ((u16)( 
(((u16)(x) & (u16)0x00ffU) << 8) | 
(((u16)(x) & (u16)0xff00U) >> 8) ));
}
			
static inline u32 _swab32(u32 x){
return ((u32)( 
(((u32)(x) & (u32)0x000000ffUL) << 24) | 
(((u32)(x) & (u32)0x0000ff00UL) <<  8) | 
(((u32)(x) & (u32)0x00ff0000UL) >>  8) | 
(((u32)(x) & (u32)0xff000000UL) >> 24) ));
}

#endif

static inline struct qnx4fs_sb_info *QNX4FS_SB(struct super_block *sb) {
	return &sb->s_fs_info;
}

static inline struct qnx4fs_inode_info *QNX4FS_I(struct inode *inode) {
	return &inode->i_info;
}

static inline unsigned long ino2phys (ino_t ino) {
	return ino >> QNX4FS_INODES_PER_BLOCK_BITS;
}

static inline int ino2slot (ino_t ino) {
	return ino & (QNX4FS_INODES_PER_BLOCK-1);
}

static inline ino_t direct2ino (unsigned long phys, int slot) {
	return ((phys - 1) << QNX4FS_INODES_PER_BLOCK_BITS) + slot;
}

static inline ino_t link2ino (qnx4fs_dirent_t *qdp) {
	return direct2ino(le32_to_cpu(qdp->l.block), qdp->l.index);
}

static inline long qnx4fs_nblocks(struct inode *ip) {
	return (ip->i_size + QNX4FS_BLOCK_SIZE - 1) >> QNX4FS_BLOCK_SIZE_BITS;
}

static inline int qnx4fs_check_range(struct super_block *sbp, unsigned long from, unsigned long len){
	unsigned long to;

	if (len == 0) return 0;
	to = from + len - 1;
	if (from <= 0 || to > QNX4FS_SB(sbp)->s_blocks)
		return -1;
	return 0;
}

extern unsigned long qnx4fs_map_block (struct inode *, long);
extern size_t strnlen (const char *, size_t);
extern int  qnx4fs_readdir (struct inode *dip, struct qnx4fs_dir *dirnt);
extern int  qnx4fs_read_inode (struct inode *);
extern int  qnx4fs_bcheck (struct super_block *sbp, unsigned long from, unsigned long len);
extern int  qnx4fs_btest_and_set (struct super_block *sbp, unsigned long from, unsigned long len);
extern int  qnx4fs_check_bitmap (struct super_block *sbp, int verbose);
extern int qnx4fs_write_new_bitmap (struct super_block *sbp);
extern int  qnx4fs_check_file (struct inode *ip);
extern void qnx4fs_print_dir (struct inode *dip, struct qnx4fs_dir *dirnt);
extern int  qnx4fs_lsblock (struct super_block*, unsigned long, int);
extern int  qnx4fs_read_file (struct inode *ip, int fd);
extern int  qnx4fs_print_inode (struct inode *ip);
extern void qnx4fs_print_xblocks (struct inode *ip);
extern int  qnx4fs_lsdir (struct inode *dip);
extern int  qnx4fs_get_devsize (struct super_block *sb);
extern int  qnx4fs_check_super (struct super_block*);
extern int  qnx4fs_find_block (struct super_block* sb, unsigned long block, char *name);

void chkfs_dump_area(const void *buf, int sze);
void chkfs_edit_block(struct super_block* sb, unsigned long block);
void chkfs_edit_dir (struct inode *dip);
void chkfs_edit_dir_blk (struct super_block* sb, unsigned long block);
void chkfs_edit_inode (struct inode *ip);
void chkfs_edit_file (struct inode *ip);
void chkfs_editor_help();

#define chkfs_EDIT   1
#define chkfs_YES    2

extern char* chkfs_mode2str (unsigned mode);
extern char* chkfs_add_name(char chr, char *str, int maxsize);
extern char* chkfs_set_name(char *str, int maxsize);
extern char* chkfs_get_name();

static inline void chkfs_restore_name(char *str){
	if(!str) return;
	*str = 0;
}

int  chkfs_check_finodes (struct inode *ip);
void chkfs_check_dir(struct inode *ip, struct inode *dip);
void chkfs_copy_dir(struct inode *ip);
void chkfs_check_links (struct super_block* sb);
struct inode *chkfs_namei(struct super_block* sb, const char *name);
void chkfs_dump(struct super_block* sb, unsigned long block);

#endif /* __QNX4FS_FS_H */
